#ifndef HIRES_H
#define HIRES_H

#include <map>
#include <fstream>
#include <sstream>
#include <string>
#include <objidl.h>
#include <gdiplus.h>
using namespace std;
using namespace Gdiplus;


struct Hires
{
	struct TexPath
	{
		string root, dir;
	};
	map<string, TexPath> texmap;

	HCRYPTPROV hCrypt;

	Hires()
	{
		hCrypt = 0;
		if (!CryptAcquireContext(&hCrypt, 0, 0, PROV_RSA_FULL, CRYPT_VERIFYCONTEXT)) hCrypt = 0;
		GdiplusStartupInput gpStartupInput;
		ULONG_PTR gpToken;
		GdiplusStartup(&gpToken, &gpStartupInput, NULL);
		loadTexMap("", "Textures\\");
	}
	~Hires()
	{
		if (hCrypt)
		{
			CryptReleaseContext(hCrypt, 0);
			hCrypt = 0;
		}
	}

	void addTexPath(const char* hex, string root, string dir)
	{
		string hash(16, 0);
		for (int i=0; i<16; i++,hex+=2)
		{
			int byte = 0;
			sscanf(hex, "%2x", &byte);
			hash[i] = (char) byte;
		}
		TexPath& tp = texmap[hash];
		tp.root = root;
		tp.dir = dir;
	}

	void loadTexMap(string root, string dir)
	{
		WIN32_FIND_DATA fd;
		HANDLE hf = FindFirstFile((dir + "*").c_str(), &fd);
		if (hf)
		{
			do
			{
				const char* name = fd.cFileName;
				int namelen = strlen(name);
				string path = dir + name;
				if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				{
					if (!strcmp(name, ".") || !strcmp(name, "..")) continue;
					path += "\\";
					string& root2 = root.empty() ? path : root;
					int num, chr;
					if (namelen == 32 && sscanf(name, "%x%c", &num, &chr) == 1) addTexPath(name, root2, path);
					else loadTexMap(root2, path);
				}
				else if (namelen > 4 && !strcmp(name + namelen - 4, ".txt"))
				{
					bool ismap = false;
					string root2 = dir, base = dir, line;
					ifstream mapfile(path, ifstream::in);
					while (getline(mapfile, line))
					{
						if (!ismap && !(ismap = line.find("GLIDOS TEXTURE MAPPING") >= 0)) continue;
						string tok1, tok2;
						istringstream(line) >> tok1 >> tok2;
						if (tok1 == "ROOT:") root2 = dir + tok2;
						if (tok1 == "BASE:") base = root2 + tok2;
						else if (tok1.size() == 32) addTexPath(tok1.c_str(), root2, base + tok2 + "\\");
					}
				}
			}
			while (FindNextFile(hf, &fd));
			FindClose(hf);
		}
	}

	void swap_red_blue(DWORD* p, int size)
	{
		while (--size >= 0)
		{
			DWORD c = *p;
			if (c < 0xFF000000) c &= 0xFFFFFF; // zbuffer does give good result with alpha blending, so avoid it
			*p++ = c & 0xFF00FF00 | c << 16 & 0xFF0000 | c >> 16 & 0xFF;
		}
	}

	bool load(PDWORD dest, PBYTE src, int wsrc, int hsrc, int scale)
	{
		if (texmap.empty()) return false;

		HCRYPTHASH hHash;
		if (!hCrypt || !CryptCreateHash(hCrypt, CALG_MD5, 0, 0, &hHash)) return false;

		BYTE hash[16];
		DWORD cbHash = sizeof(hash);
		TexPath* tp = 0;

		if (CryptHashData(hHash, src, 0x10000, 0) &&
			CryptGetHashParam(hHash, HP_HASHVAL, hash, &cbHash, 0))
		{
			auto it = texmap.find(string((char*) hash, 16));
			if (it != texmap.end()) tp = &it->second;
		}
		CryptDestroyHash(hHash);

		if (!tp) return false;

		int wdest = wsrc * scale, hdest = hsrc * scale;
		Bitmap bmpdest(wdest, hdest);
		Graphics graphics(&bmpdest);
		graphics.SetInterpolationMode(InterpolationModeHighQualityBilinear);
		graphics.SetCompositingMode(CompositingModeSourceCopy);

		// first copy original texture from dest parameter to fix incomplete hires packs
		swap_red_blue(dest, wsrc * hsrc);
		Bitmap bmpsrc(wsrc, hsrc, wsrc*4, PixelFormat32bppARGB, (PBYTE) dest);
		graphics.DrawImage(&bmpsrc, 0, 0, wdest, hdest);

		WIN32_FIND_DATA fd;
		HANDLE hf = FindFirstFile((tp->dir + "\\*").c_str(), &fd);
		if (!hf) return false;
		do
		{
			if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) continue;
			int x1,y1,x2,y2;
			if (sscanf(fd.cFileName, "(%u--%u)(%u--%u)", &x1,&x2,&y1,&y2) < 4) continue;
			x1 *= scale; y1 *= scale; x2 *= scale; y2 *= scale;
			string path = tp->dir + fd.cFileName;
			if (path.find(".txt") == path.size() - 4)
			{
				string cmd, link;
				ifstream(path, ifstream::in) >> cmd >> link;
				if (cmd == "TLNK:") path = tp->root + link;
			}
			Image frag(wstring(path.begin(), path.end()).c_str());
			graphics.DrawImage(&frag, x1, y1, x2 - x1, y2 - y1);
		}
		while (FindNextFile(hf, &fd));
		FindClose(hf);
		BitmapData bmpdata;
		bmpdata.Width = wdest;
		bmpdata.Height = hdest;
		bmpdata.Stride = 4*wdest;
		bmpdata.PixelFormat = PixelFormat32bppARGB;
		bmpdata.Scan0 = dest;
		bmpdata.Reserved = 0;
		Rect destrect(0,0,wdest,hdest);
		bmpdest.LockBits(&destrect, ImageLockModeRead|ImageLockModeUserInputBuf, PixelFormat32bppARGB, &bmpdata);
		bmpdest.UnlockBits(&bmpdata);
		swap_red_blue(dest, wdest * hdest);
		return true;
	}
};

#endif